@@ -2,6 +2,8 @@ import datetime as dt
 import pickle
 import socket
 import subprocess
+import sys
+import time
 from contextlib import suppress
 from pathlib import Path
 
@@ -11,8 +13,26 @@ from mbedtls import hashlib
 from mbedtls._tls import _DTLSCookie as DTLSCookie
 from mbedtls._tls import _PSKSToreProxy as PSKStoreProxy
 from mbedtls.pk import RSA
-from mbedtls.tls import *
-from mbedtls.tls import HandshakeStep, TLSSession
+from mbedtls.tls import (
+    ClientContext,
+    DTLSConfiguration,
+    DTLSVersion,
+    HandshakeStep,
+    HelloVerifyRequest,
+    NextProtocol,
+    Purpose,
+    ServerContext,
+    TLSConfiguration,
+    TLSRecordHeader,
+    TLSSession,
+    TLSVersion,
+    TLSWrappedBuffer,
+    TLSWrappedSocket,
+    TrustStore,
+    WantReadError,
+    WantWriteError,
+    ciphers_available,
+)
 from mbedtls.x509 import CRT, CSR, BasicConstraints
 
 
@@ -21,6 +41,77 @@ def rootpath():
     return Path(__file__).parent.parent
 
 
+def make_root_ca(
+    subject=None,
+    not_before=None,
+    not_after=None,
+    serial_number=None,
+    basic_constraints=None,
+    digestmod=None,
+):
+    if subject is None:
+        subject = "OU=test, CN=Trusted CA"
+    if not_before is None:
+        not_before = dt.datetime.utcnow()
+    if not_after is None:
+        not_after = not_before + dt.timedelta(days=90)
+    if serial_number is None:
+        serial_number = 0x123456
+    if basic_constraints is None:
+        basic_constraints = BasicConstraints(True, -1)
+    if digestmod is None:
+        digestmod = hashlib.sha256
+
+    key = RSA()
+    key.generate()
+    crt = CRT.selfsign(
+        csr=CSR.new(key, subject, digestmod()),
+        issuer_key=key,
+        not_before=not_before,
+        not_after=not_after,
+        serial_number=serial_number,
+        basic_constraints=basic_constraints,
+    )
+    return crt, key
+
+
+def make_crt(
+    issuer_crt,
+    issuer_key,
+    subject=None,
+    not_before=None,
+    not_after=None,
+    serial_number=None,
+    basic_constraints=None,
+    digestmod=None,
+):
+    if subject is None:
+        subject = "OU=test, CN=hostname"
+    if not_before is None:
+        not_before = issuer_crt.not_before
+    if not_after is None:
+        not_after = issuer_crt.not_after
+    if serial_number is None:
+        serial_number = 0x123456
+    if basic_constraints is None:
+        basic_constraints = BasicConstraints()
+    if digestmod is None:
+        # TODO: issuer_crt.digestmod should work but doesn't.
+        digestmod = hashlib.sha256
+
+    key = RSA()
+    key.generate()
+    crt = issuer_crt.sign(
+        csr=CSR.new(key, subject, digestmod()),
+        issuer_key=issuer_key,
+        not_before=not_before,
+        not_after=not_after,
+        serial_number=serial_number,
+        basic_constraints=basic_constraints,
+    )
+    return crt, key
+
+
 class TestPickle:
     @pytest.mark.parametrize(
         "obj",
@@ -190,70 +281,7 @@ class TestTLSSession:
         assert isinstance(repr(session), str)
 
 
-class Chain:
-    @pytest.fixture(scope="class")
-    def now(self):
-        return dt.datetime.utcnow()
-
-    @pytest.fixture(scope="class")
-    def digestmod(self):
-        return hashlib.sha256
-
-    @pytest.fixture(scope="class")
-    def ca0_key(self):
-        ca0_key = RSA()
-        ca0_key.generate()
-        return ca0_key
-
-    @pytest.fixture(scope="class")
-    def ca1_key(self):
-        ca1_key = RSA()
-        ca1_key.generate()
-        return ca1_key
-
-    @pytest.fixture(scope="class")
-    def ee0_key(self):
-        ee0_key = RSA()
-        ee0_key.generate()
-        return ee0_key
-
-    @pytest.fixture(scope="class")
-    def ca0_crt(self, ca0_key, digestmod, now):
-        ca0_csr = CSR.new(ca0_key, "CN=Trusted CA", digestmod())
-        return CRT.selfsign(
-            ca0_csr,
-            ca0_key,
-            not_before=now,
-            not_after=now + dt.timedelta(days=90),
-            serial_number=0x123456,
-            basic_constraints=BasicConstraints(True, -1),
-        )
-
-    @pytest.fixture(scope="class")
-    def ca1_crt(self, ca1_key, ca0_crt, ca0_key, digestmod, now):
-        ca1_csr = CSR.new(ca1_key, "CN=Intermediate CA", digestmod())
-        return ca0_crt.sign(
-            ca1_csr,
-            ca0_key,
-            now,
-            now + dt.timedelta(days=90),
-            0x234567,
-            basic_constraints=BasicConstraints(True, -1),
-        )
-
-    @pytest.fixture(scope="class")
-    def ee0_crt(self, ee0_key, ca1_crt, ca1_key, digestmod, now):
-        ee0_csr = CSR.new(ee0_key, "CN=End Entity", digestmod())
-        return ca1_crt.sign(
-            ee0_csr, ca1_key, now, now + dt.timedelta(days=90), 0x345678
-        )
-
-    @pytest.fixture(scope="class")
-    def certificate_chain(self, ee0_crt, ca1_crt, ee0_key):
-        return (ee0_crt, ca1_crt), ee0_key
-
-
-class TestTrustStore(Chain):
+class TestTrustStore:
     @pytest.fixture
     def store(self):
         return TrustStore.system()
@@ -285,9 +313,10 @@ class TestTrustStore(Chain):
         store.add(store[0])
         assert len(store) == length
 
-    def test_add_new_certificate(self, store, ca0_crt):
+    def test_add_new_certificate(self, store):
+        root_ca = make_root_ca()[0]
         length = len(store)
-        store.add(ca0_crt)
+        store.add(root_ca)
         assert len(store) == length + 1
 
 
@@ -305,7 +334,7 @@ class TestDTLSCookie:
         assert cookie.timeout == 1000
 
 
-class _BaseConfiguration(Chain):
+class _BaseConfiguration:
     @pytest.fixture
     def conf(self):
         raise NotImplementedError
@@ -324,9 +353,11 @@ class _BaseConfiguration(Chain):
         assert conf_.validate_certificates is validate
 
     @pytest.mark.parametrize("chain", [((), None), None])
-    def test_set_certificate_chain(self, conf, chain, certificate_chain):
+    def test_set_certificate_chain(self, conf, chain):
+        root_crt, root_key = make_root_ca()
+        ee_crt, ee_key = make_crt(root_crt, root_key)
         if chain is None:
-            chain = certificate_chain
+            chain = (ee_crt, root_crt), ee_key
         conf_ = conf.update(certificate_chain=chain)
         assert conf_.certificate_chain == chain
 
@@ -491,25 +522,44 @@ SERVER_HELLO = (
     HandshakeStep.SERVER_HELLO_DONE,
 )
 CLIENT_KEY_EXCHANGE = (
+    HandshakeStep.CLIENT_CERTIFICATE,
     HandshakeStep.CLIENT_KEY_EXCHANGE,
     HandshakeStep.CERTIFICATE_VERIFY,
     HandshakeStep.CLIENT_CHANGE_CIPHER_SPEC,
     HandshakeStep.CLIENT_FINISHED,
 )
-CHANGE_CIPHER_SPEC = (
+SERVER_CHANGE_CIPHER_SPEC = (
     HandshakeStep.SERVER_CHANGE_CIPHER_SPEC,
     HandshakeStep.SERVER_FINISHED,
+)
+HANDSHAKE_OVER = (
     HandshakeStep.FLUSH_BUFFERS,
     HandshakeStep.HANDSHAKE_WRAPUP,
     HandshakeStep.HANDSHAKE_OVER,
 )
 
 
+def make_server(conf):
+    return ServerContext(conf).wrap_buffers()
+
+
+def make_client(conf, hostname):
+    return ClientContext(conf).wrap_buffers(hostname)
+
+
 def do_io(*, src, dst, amt=1024):
     __tracebackhide__ = True
-    in_transit = src.peek_outgoing(amt)
-    src.consume_outgoing(len(in_transit))
-    dst.receive_from_network(in_transit)
+    assert src._output_buffer, "nothing to do"
+    while src._output_buffer:
+        in_transit = src.peek_outgoing(amt)
+        src.consume_outgoing(len(in_transit))
+        dst.receive_from_network(in_transit)
+
+
+def do_send(data, *, src, dst):
+    amt = src.write(data)
+    do_io(src=src, dst=dst)
+    return dst.read(amt)
 
 
 def do_handshake(end, states):
@@ -517,30 +567,38 @@ def do_handshake(end, states):
     while end._handshake_state is not states[0]:
         # The backend goes through every state for both
         # ends.  This is not relevant.
-        end.do_handshake()
+        try:
+            end.do_handshake()
+        except (WantReadError, WantWriteError) as exc:
+            raise AssertionError(
+                f"{type(end.context).__name__} wants {end._handshake_state}"
+            ) from exc
 
     for state in states:
         assert end._handshake_state is state
         if state is HandshakeStep.HANDSHAKE_OVER:
             break
-        with suppress(WantReadError, WantWriteError):
+        with suppress(WantWriteError):
             end.do_handshake()
 
 
 def make_full_handshake(*, client, server):
     do_handshake(client, CLIENT_HELLO)
     do_io(src=client, dst=server)
+
     do_handshake(server, SERVER_HELLO)
+    do_io(src=server, dst=client)
     assert client.negotiated_protocol() == server.negotiated_protocol()
 
-    do_io(src=server, dst=client)
     do_handshake(client, CLIENT_KEY_EXCHANGE)
+    do_io(src=client, dst=server)
     assert client.negotiated_tls_version() == server.negotiated_tls_version()
 
-    do_io(src=client, dst=server)
-    do_handshake(server, CHANGE_CIPHER_SPEC)
+    do_handshake(server, SERVER_CHANGE_CIPHER_SPEC)
     do_io(src=server, dst=client)
-    do_handshake(client, (HandshakeStep.HANDSHAKE_OVER,))
+
+    do_handshake(server, HANDSHAKE_OVER)
+    do_handshake(client, HANDSHAKE_OVER)
     assert client.cipher() == server.cipher()
 
 
@@ -562,72 +620,150 @@ def make_hello_verify_request(*, client, server, cookie):
     do_io(src=server, dst=client)
 
 
-class TestHandshake:
-    def test_tls(self):
-        psk = ("cli", b"secret")
+def do_communicate(args):
+    while True:
+        with subprocess.Popen(
+            args,
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+            text=True,
+            encoding="utf8",
+        ) as proc:
+            out, err = proc.communicate()
+            if "ConnectionRefusedError" not in err:
+                return out
+            time.sleep(0.01)  # Avoid tight CPU loop.
+            continue
 
-        srv_ctx = ServerContext(
+
+class TestTLSHandshake:
+    @pytest.fixture(scope="class")
+    def hostname(self):
+        return "www.example.com"
+
+    @pytest.fixture(scope="class")
+    def certificate_chain(self, hostname):
+        root_crt, root_key = make_root_ca()
+        ee_crt, ee_key = make_crt(
+            root_crt, root_key, subject=f"OU=test, CN={hostname}"
+        )
+        return (ee_crt, root_crt), ee_key
+
+    def test_cert_without_validation(self, certificate_chain):
+        server = make_server(
             TLSConfiguration(
-                pre_shared_key_store=dict((psk,)),
+                certificate_chain=certificate_chain,
                 validate_certificates=False,
             )
         )
-        cli_ctx = ClientContext(
-            TLSConfiguration(
-                pre_shared_key=psk,
-                validate_certificates=False,
-            )
+        client = make_client(
+            TLSConfiguration(validate_certificates=False), "hostname"
         )
-        server = srv_ctx.wrap_buffers()
-        client = cli_ctx.wrap_buffers("hostname")
-
         make_full_handshake(client=client, server=server)
 
-        secret = b"a very secret message"
+        secret = "a very secret message".encode("utf8")
+        assert do_send(secret, src=client, dst=server) == secret
+        assert do_send(secret, src=server, dst=client) == secret
 
-        amt = client.write(secret)
-        do_io(src=client, dst=server)
-        assert server.read(amt) == secret
+    def test_cert_with_validation(self, hostname, certificate_chain):
+        trust_store = TrustStore()
+        for crt in certificate_chain[0][1:]:
+            trust_store.add(crt)
+        server = make_server(
+            TLSConfiguration(
+                certificate_chain=certificate_chain,
+                validate_certificates=False,
+            )
+        )
+        # Host name must now be the common name (CN) of the leaf certificate.
+        client = make_client(
+            TLSConfiguration(trust_store=trust_store), hostname
+        )
+        make_full_handshake(client=client, server=server)
 
-        amt = server.write(secret)
-        do_io(src=server, dst=client)
-        assert client.read(amt) == secret
+        secret = "a very secret message".encode("utf8")
+        assert do_send(secret, src=client, dst=server) == secret
+        assert do_send(secret, src=server, dst=client) == secret
 
-    def test_dtls(self):
+    def test_psk(self):
         psk = ("cli", b"secret")
+        server = make_server(
+            TLSConfiguration(
+                pre_shared_key_store=dict((psk,)),
+                validate_certificates=False,
+            )
+        )
+        client = make_client(
+            TLSConfiguration(
+                pre_shared_key=psk,
+                validate_certificates=False,
+            ),
+            "hostname",
+        )
+        make_full_handshake(client=client, server=server)
 
-        srv_ctx = ServerContext(
+        secret = "a very secret message".encode("utf8")
+        assert do_send(secret, src=client, dst=server) == secret
+        assert do_send(secret, src=server, dst=client) == secret
+
+
+class TestDTLSHandshake:
+    def test_psk(self):
+        psk = ("cli", b"secret")
+        server = make_server(
             DTLSConfiguration(
                 pre_shared_key_store=dict((psk,)),
                 validate_certificates=False,
             )
         )
-        cli_ctx = ClientContext(
+        client = make_client(
             DTLSConfiguration(
                 pre_shared_key=psk,
                 validate_certificates=False,
-            )
+            ),
+            "hostname",
         )
-        server = srv_ctx.wrap_buffers()
-        client = cli_ctx.wrap_buffers("hostname")
-
         make_hello_verify_request(
             client=client, server=server, cookie="🍪🍪🍪".encode("utf-8")
         )
         make_full_handshake(client=client, server=server)
 
-        secret = b"a very secret message"
+        secret = "a very secret message".encode("utf8")
+        assert do_send(secret, src=client, dst=server) == secret
+        assert do_send(secret, src=server, dst=client) == secret
 
-        amt = client.write(secret)
-        do_io(src=client, dst=server)
-        assert server.read(amt) == secret
+    def test_resume_from_pickle(self):
+        psk = ("cli", b"secret")
+        server = make_server(
+            DTLSConfiguration(
+                pre_shared_key_store=dict((psk,)),
+                validate_certificates=False,
+            )
+        )
+        client = make_client(
+            DTLSConfiguration(
+                pre_shared_key=psk,
+                ciphers=["TLS-ECDHE-PSK-WITH-CHACHA20-POLY1305-SHA256"],
+                lowest_supported_version=DTLSVersion.DTLSv1_2,
+                validate_certificates=False,
+            ),
+            "hostname",
+        )
+        make_hello_verify_request(
+            client=client, server=server, cookie="🍪🍪🍪".encode("utf-8")
+        )
+        make_full_handshake(client=client, server=server)
 
-        amt = server.write(secret)
-        do_io(src=server, dst=client)
-        assert client.read(amt) == secret
+        secret = "a very secret message".encode("utf8")
+        do_send(secret, src=client, dst=server)
+        do_send(secret, src=server, dst=client)
+
+        client = pickle.loads(pickle.dumps(client))
+        do_send(secret, src=client, dst=server)
+        do_send(secret, src=server, dst=client)
 
 
-@pytest.mark.local
+@pytest.mark.skipif(sys.platform == "win32", reason="Flaky under Windows")
 class TestProgramsTLS:
     @pytest.fixture
     def port(self):
@@ -638,22 +774,23 @@ class TestProgramsTLS:
 
         """
         port = 0
-        with socket.socket() as sock:
-            sock.bind(("", port))
+        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
+            sock.bind(("localhost", port))
             port = sock.getsockname()[1]
         return port
 
     @pytest.fixture
     def server(self, rootpath, port):
         args = [
-            rootpath / "programs" / "server.py",
+            sys.executable,
+            str(rootpath / "programs" / "server.py"),
             "--port",
             f"{port}",
             "--tls",
             "--psk-store",
             "cli=secret",
         ]
-        proc = subprocess.Popen(args)
+        proc = subprocess.Popen(args, text=True, encoding="utf8")
         yield proc
         proc.kill()
         proc.wait(1.0)
@@ -661,9 +798,10 @@ class TestProgramsTLS:
     @pytest.mark.usefixtures("server")
     @pytest.mark.timeout(10)
     def test_communicate(self, rootpath, port):
-        secret = b"a very secret message"
+        secret = "a very secret message"
         args = [
-            rootpath / "programs" / "client.py",
+            sys.executable,
+            str(rootpath / "programs" / "client.py"),
             "--port",
             f"{port}",
             "--tls",
@@ -672,12 +810,10 @@ class TestProgramsTLS:
             secret,
         ]
         for _ in range(3):
-            with subprocess.Popen(args, stdout=subprocess.PIPE) as client:
-                out, err = client.communicate()
-                assert out == secret + b"\n"
+            assert do_communicate(args) == secret + "\n"
 
 
-@pytest.mark.local
+@pytest.mark.skipif(sys.platform == "win32", reason="Flaky under Windows")
 class TestProgramsDTLS:
     @pytest.fixture
     def port(self):
@@ -688,22 +824,23 @@ class TestProgramsDTLS:
 
         """
         port = 0
-        with socket.socket() as sock:
-            sock.bind(("", port))
+        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
+            sock.bind(("localhost", port))
             port = sock.getsockname()[1]
         return port
 
     @pytest.fixture
     def server(self, rootpath, port):
         args = [
-            rootpath / "programs" / "server.py",
+            sys.executable,
+            str(rootpath / "programs" / "server.py"),
             "--port",
             f"{port}",
             "--dtls",
             "--psk-store",
             "cli=secret",
         ]
-        proc = subprocess.Popen(args)
+        proc = subprocess.Popen(args, text=True, encoding="utf8")
         yield proc
         proc.kill()
         proc.wait(1.0)
@@ -711,9 +848,10 @@ class TestProgramsDTLS:
     @pytest.mark.usefixtures("server")
     @pytest.mark.timeout(10)
     def test_communication(self, rootpath, port):
-        secret = b"a very secret message"
+        secret = "a very secret message"
         args = [
-            rootpath / "programs" / "client.py",
+            sys.executable,
+            str(rootpath / "programs" / "client.py"),
             "--port",
             f"{port}",
             "--dtls",
@@ -722,6 +860,4 @@ class TestProgramsDTLS:
             secret,
         ]
         for _ in range(3):
-            with subprocess.Popen(args, stdout=subprocess.PIPE) as client:
-                out, err = client.communicate()
-                assert out == secret + b"\n"
+            assert do_communicate(args) == secret + "\n"
